#
#  Copyright 1999-2005 Illinois Institute of Technology
#
#  See the file "LICENSE" for information on usage and redistribution
#  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#

proc setup_imcagui_callbacks { record_list } {
	global callback_records

	find_callback_record $record_list "lambda"       "lambda"
	find_callback_record $record_list "d_spacing"    "d_spacing"

	if { $IMCAGui::show_slit_positions } {
		find_callback_record $record_list "hslit_center" "bh_center"
		find_callback_record $record_list "vslit_center" "bv_center"
	}

	if { $IMCAGui::show_ring_current } {
		find_callback_record $record_list "ring_current" "ring_current"
	}

	if { $IMCAGui::show_ion_chamber } {
		find_callback_record $record_list "ion_chamber" "scaler2"
	}

	find_callback_record $record_list "hslit_size"   "bh_size"
	find_callback_record $record_list "vslit_size"   "bv_size"
	find_callback_record $record_list "edge_energy"  "edge_energy"

	if { $IMCAGui::have_undulator } {
	    find_callback_record $record_list "id_ev"          "id_ev"
	    find_callback_record $record_list "id_ev_harmonic" "id_ev_harmonic"
	    find_callback_record $record_list "id_ev_params"   "id_ev_params"
	    find_callback_record $record_list "id_ev_enabled"  "id_ev_enabled"
	}

	if { $IMCAGui::have_filters } {
	    if { $IMCAGui::beamline_name == "imca17id" } {
	        find_callback_record $record_list "filter1"    "filter1"
	        find_callback_record $record_list "filter2"    "filter2"
	    } else {
		find_callback_record $record_list "pfcu1"      "pfcu1"
		find_callback_record $record_list "pfcu2"      "pfcu2"
	    }
	}

	if { $IMCAGui::have_bpms } {
	    find_callback_record $record_list "bpm_x" "bpm_ds_posx"
	    find_callback_record $record_list "bpm_y" "bpm_ds_posy"
	}

	if { $IMCAGui::beamline_name == "imca17bm" } {
		find_callback_record $record_list \
					"bshutter_closed" "bshutter_closed"
		find_callback_record $record_list \
					"bshutter_open" "bshutter_open"
		find_callback_record $record_list "x_fb_enable" "x_fb_enable"
		find_callback_record $record_list "y_fb_enable" "y_fb_enable"

		set Imca17bmFeedback::x_fb_enable $callback_records(x_fb_enable)
		set Imca17bmFeedback::y_fb_enable $callback_records(y_fb_enable)

		set Imca17bmFeedback::horiz_motor \
			[ $record_list get_record "chi" ]
		set Imca17bmFeedback::vert_motor \
			[ $record_list get_record "mjack_pitch" ]

		find_callback_record $record_list "focus" "focus"
	}

	start_imcagui_callbacks
}

############################################################################

namespace eval IMCAGui {
	# The array edge_array stores information about the available
	# element and edge combinations as read from 'elements.dat'
	#

	variable edge_array
}

if { $beamline_name == "imca17id" } {

	namespace eval IMCAGui {
		variable program_label "IMCA-CAT Sector 17-ID"

		variable mca_name "xflash"

		variable minimum_edge_energy  7000
		variable maximum_edge_energy 14500

		variable show_ring_current 1

		variable have_undulator 1

		variable have_filters 1

		variable show_slit_positions 0

		variable show_ion_chamber 0

		variable have_bpms 0

		variable width 780
		variable height 680

		variable motor_name_to_field_name

		variable filter1_thickness
		variable filter2_thickness
		variable filter1_position
		variable filter2_position
		variable thickness_list
		variable thickness_indices

		# Foil thickness is in micrometers.

		set foil_thickness 25.0

		set filter1_thickness(0)     0
		set filter1_thickness(1)     1
		set filter1_thickness(2)     7
		set filter1_thickness(3)    16
		set filter1_thickness(4)    25

		set filter1_position(0)      0
		set filter1_position(1)  16800
		set filter1_position(2)  33500
		set filter1_position(3)  50700
		set filter1_position(4)  67450

		set filter2_thickness(0)     0
		set filter2_thickness(1)     2
		set filter2_thickness(2)     4
		set filter2_thickness(3)    14
		set filter2_thickness(4)    40

		set filter2_position(0)      0
		set filter2_position(1)  16900
		set filter2_position(2)  33900
		set filter2_position(3)  50900
		set filter2_position(4)  67775

		set nelements1 [ array size filter1_thickness ]
		set nelements2 [ array size filter2_thickness ]

		# Multiply the number of foil thicknesses by the thickness
		# of an individual foil.

		for { set i 0 } { $i < $nelements1 } { incr i } {
			set filter1_thickness($i) \
			[ expr ( $filter1_thickness($i) * $foil_thickness ) ]
		}
		for { set i 0 } { $i < $nelements2 } { incr i } {
			set filter2_thickness($i) \
			[ expr ( $filter2_thickness($i) * $foil_thickness ) ]
		}

		# Make a list of all the foil thickness combinations.

		for { set i 0 } { $i < $nelements1 } { incr i } {
			for { set j 0 } { $j < $nelements2 } { incr j } {

				set t1 $filter1_thickness($i)
				set t2 $filter2_thickness($j)

				lappend thickness_list [ expr ($t1 + $t2) ]

				lappend thickness_indices [ list $i $j ]
			}
		}
	}

} else {
	;# IMCA Sector 17-BM

	source "$env(MXDIR)/etc/imca17bm_feedback_interlock.tcl"

	namespace eval IMCAGui {
		variable program_label "IMCA-CAT Sector 17-BM"

		variable minimum_edge_energy  5920
		variable maximum_edge_energy 14000

		variable show_ring_current 1

		variable have_undulator 0

		variable have_filters 1

		variable show_slit_positions 0

		variable show_ion_chamber 1

		variable have_bpms 1

		variable width 680
		variable height 750

		variable motor_name_to_field_name

		variable filter1_thickness
		variable filter2_thickness
		variable thickness_list

		variable maximum_filter_thickness

		# Foil thickness is in micrometers.

		set foil_thickness 25.0

		set filter1_thickness(0)     1
		set filter1_thickness(1)     2
		set filter1_thickness(2)     4
		set filter1_thickness(3)     8

		set filter2_thickness(0)    12
		set filter2_thickness(1)    16
		set filter2_thickness(2)    20
		set filter2_thickness(3)    40

		set nelements1 [ array size filter1_thickness ]
		set nelements2 [ array size filter2_thickness ]

		# Multiply the number of foil thicknesses by the thickness
		# of an individual foil.

		for { set i 0 } { $i < $nelements1 } { incr i } {
			set filter1_thickness($i) \
			[ expr ( $filter1_thickness($i) * $foil_thickness ) ]
		}
		for { set i 0 } { $i < $nelements2 } { incr i } {
			set filter2_thickness($i) \
			[ expr ( $filter2_thickness($i) * $foil_thickness ) ]
		}

		# Compute the maximum filter thickness.

		set maximum_filter_thickness 0

		for { set i 0 } { $i < $nelements1 } { incr i } {
			set maximum_filter_thickness \
		 [ expr ( $maximum_filter_thickness + $filter1_thickness($i) ) ]
		}
		for { set i 0 } { $i < $nelements2 } { incr i } {
			set maximum_filter_thickness \
		 [ expr ( $maximum_filter_thickness + $filter2_thickness($i) ) ]
		}

		# Make a list of all the foil thickness combinations

		set foil_thickness_list [ list ]

		for { set i 0 } { $i < $nelements1 } { incr i } {
			lappend foil_thickness_list $filter1_thickness($i)
		}

		for { set i 0 } { $i < $nelements2 } { incr i } {
			lappend foil_thickness_list $filter2_thickness($i)
		}

		set foil_list_length [ llength $foil_thickness_list ]

		set max_bitmask [ expr ( pow( 2, $foil_list_length ) - 1 ) ]

		set thickness_list [ list ]

		for { set i 0 } { $i < $max_bitmask} { incr i } {
			set total_thickness 0

			for { set j 0 } { $j < $foil_list_length } { incr j } {
				set bit_is_set [ expr ( (1 << $j) & $i ) ]

				if { $bit_is_set } {
					set item_thickness \
					    [ lindex $foil_thickness_list $j ]

					set total_thickness \
			    [ expr ( $total_thickness + $item_thickness ) ]
				}
			}

			lappend thickness_list $total_thickness
		}

#		log_puts "thickness_list = '$thickness_list'"
	}
}

############################################################################

proc create_bpm_frame { record_list frame } {

	package require BLT

	set bpm_plot $frame.bpm_plot

	set IMCAGui::bpm_plot $bpm_plot

	blt::graph $bpm_plot -plotbackground white

	$bpm_plot element create beamspot

	# Configure the BPM plot

	$bpm_plot configure -title "Beam Position Monitor"

	$bpm_plot element configure beamspot -label ""
	$bpm_plot element configure beamspot -color red

	$bpm_plot axis configure x -min -4000.0 -max 4000.0
	$bpm_plot axis configure y -min -4000.0 -max 4000.0

	$bpm_plot axis configure x -title "x (um)"
	$bpm_plot axis configure y -title "y (um)"

	$bpm_plot grid configure -hide 0

	pack $bpm_plot -side left

	# Display the x and y positions

	set xyframe $frame.xyframe

	frame $xyframe
	pack $xyframe -side left

	label $xyframe.x_label -text "X:"
	label $xyframe.x_value -text "----" -fg blue -width 12
	label $xyframe.x_units -text "um"

	grid $xyframe.x_label -row 0 -column 0
	grid $xyframe.x_value -row 0 -column 1
	grid $xyframe.x_units -row 0 -column 2

	label $xyframe.y_label -text "Y:"
	label $xyframe.y_value -text "----" -fg blue -width 12
	label $xyframe.y_units -text "um"

	grid $xyframe.y_label -row 1 -column 0
	grid $xyframe.y_value -row 1 -column 1
	grid $xyframe.y_units -row 1 -column 2
}

############################################################################

proc optimize_menu_handler { frame } {

	# If the menu entries have already been created, just return.

	if [ winfo exists $frame.optimize.tag ] {
		return
	}

	label $frame.optimize.tag	;# Do not pack or grid this.

	set menu $frame.optimize.menu

	$menu add command -label "Manual Optimize..." \
		-command [ itcl::code eval exec "optimize &" ]

	# JLM, 2004-05-27: Commented out this menu since it does not do
	# anything right now and the optimize_staff command does not exist.
	# This change was OKed by Bill.
	#$menu add command -label "Manual Optimize (staff)..." \
	#	-command [ itcl::code eval exec "optimize_staff &" ]

	$menu add command -label "Auto Optimize Monochromator..." \
		-command [ itcl::code eval exec "optimize_auto_mono &" ]

	$menu add command -label "Auto Optimize Slits..." \
		-command [ itcl::code eval exec "optimize_auto_slits &" ]

	if { $IMCAGui::beamline_name == "imca17id" } {
		$menu add command -label "Auto Optimize Table..." \
		-command [ itcl::code eval exec "optimize_auto_table &" ]
	}

	$menu add command -label "Auto Optimize All..." \
		-command [ itcl::code eval exec "optimize_auto_all &" ]
}

############################################################################

proc move_17id_filters { record_list grid } {

	set filter1 [ $record_list get_record "filter1" ]
	set filter2 [ $record_list get_record "filter2" ]

	set thickness_desired [ $grid.filter_desired get ]

	set i [ lsearch -exact $IMCAGui::thickness_list $thickness_desired ]

	if { $i < 0 } {
		set t "Attenuator thickness $thickness_desired not found in "
		append t "filter thickness list."

		tk_dialog .bad_filter_thickness_dialog \
		"Bad Attenuator Thickness" $t error 0 "OK"

		return
	}

	set index_list [ lindex $IMCAGui::thickness_indices $i ]

	set index1 [ lindex $index_list 0 ]
	set index2 [ lindex $index_list 1 ]

	set filter1_position $IMCAGui::filter1_position($index1)
	set filter2_position $IMCAGui::filter2_position($index2)

	set button $grid.move

	set_move_state_to_moving $button

	$filter1 move_absolute $filter1_position
	$filter2 move_absolute $filter2_position
}

############################################################################

proc get_17id_filter_thickness {} {

	global callback_values
	global callback_records

	if [ $callback_records(filter1) is_busy ] {
		return "Moving"
	}
	if [ $callback_records(filter2) is_busy ] {
		return "Moving"
	}

	set threshold 10

	set position1 $callback_values(filter1)
	set position2 $callback_values(filter2)

	set thickness1 -1
	set thickness2 -1

	set nelements [ array size IMCAGui::filter1_position ]

	for { set i 0 } { $i < $nelements } { incr i } {

		set item_position $IMCAGui::filter1_position($i)

		set position_diff \
			[ expr ( abs( $item_position - $position1 ) ) ]

		if { $position_diff < $threshold } {
			set thickness1 $IMCAGui::filter1_thickness($i)
			break
		}
	}

	if { $thickness1 < 0 } {
		return "Fault"
	}

	set nelements [ array size IMCAGui::filter2_position ]

	for { set i 0 } { $i < $nelements } { incr i } {

		set item_position $IMCAGui::filter2_position($i)

		set position_diff \
			[ expr ( abs( $item_position - $position2 ) ) ]

		if { $position_diff < $threshold } {
			set thickness2 $IMCAGui::filter2_thickness($i)
			break
		}
	}

	if { $thickness2 < 0 } {
		return "Fault"
	}

	set result [ expr ( $thickness1 + $thickness2 ) ]

	return $result
}

############################################################################

proc get_17bm_filter_thickness {} {

	global callback_values
	global callback_records

	set pfcu1 $callback_values(pfcu1)
	set pfcu2 $callback_values(pfcu2)

	set total_foils 0

	for { set i 0 } { $i < 4 } { incr i } {
		set thickness $IMCAGui::filter1_thickness($i)

		if { $pfcu1 & ( 1 << $i ) } {
			set total_foils [ expr ( $total_foils + $thickness ) ]
		}
	}

	for { set i 0 } { $i < 4 } { incr i } {
		set thickness $IMCAGui::filter2_thickness($i)

		if { $pfcu2 & ( 1 << $i ) } {
			set total_foils [ expr ( $total_foils + $thickness ) ]
		}
	}

	return $total_foils
}

############################################################################

proc change_17bm_filters { record_list grid } {

	set new_thickness [ $grid.filter_desired get ]

	if { $new_thickness == "" } {
		set new_thickness 0
	}

#	log_puts "new_thickness = $new_thickness"

# Figure out which filters to insert.  This logic implicitly _assumes_ that
# the filter thicknesses are arranged in a monotonically decreasing order.

	set pfcu1_value 0
	set pfcu2_value 0

	set remainder $new_thickness

# First look at filter 2.

	for { set i 3 } { $i >= 0 } { incr i -1 } {
		set filter_thickness $IMCAGui::filter2_thickness($i)

#		log_puts \
#  "#2, i = $i, filter_thickness = $filter_thickness, remainder = $remainder"

		if { $filter_thickness <= $remainder } {
			set pfcu2_value [ expr ( $pfcu2_value | ( 1 << $i ) ) ]

			set remainder [expr ( $remainder - $filter_thickness )]
		}
	}

# Then look at filter 1.

	for { set i 3 } { $i >= 0 } { incr i -1 } {
		set filter_thickness $IMCAGui::filter1_thickness($i)

#		log_puts \
#  "#1, i = $i, filter_thickness = $filter_thickness, remainder = $remainder"

		if { $filter_thickness <= $remainder } {
			set pfcu1_value [ expr ( $pfcu1_value | ( 1 << $i ) ) ]

			set remainder [expr ( $remainder - $filter_thickness )]
		}
	}

#	log_puts "pfcu2_value = $pfcu2_value, pfcu1_value = $pfcu1_value"

	set pfcu2 [ $record_list get_record "pfcu2" ]

	set pfcu1 [ $record_list get_record "pfcu1" ]

	$pfcu2 write $pfcu2_value

	$pfcu1 write $pfcu1_value
}

############################################################################

proc create_filter_frame { frame  record_list } {

	if { $IMCAGui::beamline_name == "imca17id" } {
		create_17id_filter_frame $frame $record_list
	} else {
		create_17bm_filter_frame $frame $record_list
	}
}

############################################################################

proc create_17id_filter_frame { frame  record_list } {

        set grid $frame.grid

        frame $grid
        pack $grid -side top

        # Create the filter control grid top labels.

	label $grid.frame_label -text "Attenuation:  "
        label $grid.current -text "Current"
        label $grid.desired -text "Desired"

	grid $grid.frame_label  -row 0 -column 0
        grid $grid.current -row 0 -column 1 -columnspan 2
        grid $grid.desired -row 0 -column 3 -columnspan 2

        # Create the filter row.

        label $grid.filter_label -text "Al Thickness "
        label $grid.filter_current -fg blue -text "---"
        label $grid.filter_current_units -text "um  "        

	iwidgets::optionmenu $grid.filter_desired

        set sorted_thickness_list [ lsort -real $IMCAGui::thickness_list ]

        foreach item $sorted_thickness_list {
                $grid.filter_desired insert end $item
        }

#        set thickness [ get_17id_filter_thickness ]

        set thickness 0

        if { $thickness == "Moving" } {
                set thickness 0
        }
        if { $thickness == "Fault" } {
                set thickness 0
        }
        set selection [ lsearch -exact $sorted_thickness_list $thickness ]

        if { $selection < 0 } {
                set selection 0
        }

        $grid.filter_desired select $selection

        button $grid.move -text "Move" -width 6 \
                -command [ itcl::code move_17id_filters $record_list $grid ]

	$grid.move configure -disabledforeground \
			[ $grid.move cget -foreground ]

        label $grid.filter_desired_units -text "um  "

        grid $grid.filter_label         -row 1 -column 0
        grid $grid.filter_current       -row 1 -column 1
        grid $grid.filter_current_units -row 1 -column 2 -sticky w
        grid $grid.filter_desired       -row 1 -column 3
        grid $grid.filter_desired_units -row 1 -column 4 -sticky w
        grid $grid.move                 -row 1 -column 5

        # Create the bottom row.

        #set bottom $frame.bottom
        set bottom $frame

        #frame $bottom
        #pack $bottom -side top

#        pack $bottom.move -side right
}

############################################################################

proc create_17bm_filter_frame { frame  record_list } {

        set grid $frame.grid

        frame $grid
        pack $grid -side top

        # Create the filter control grid top labels.

	label $grid.frame_label -text "Attenuation:  "
        label $grid.current -text "Current"
        label $grid.desired -text "Desired"

	grid $grid.frame_label  -row 0 -column 0
        grid $grid.current -row 0 -column 1 -columnspan 2
        grid $grid.desired -row 0 -column 3 -columnspan 2

        # Create the filter row.

        label $grid.filter_label -text "Al Thickness "
        label $grid.filter_current -fg blue -text "---"
        label $grid.filter_current_units -text "um  "        

	iwidgets::combobox $grid.filter_desired \
		-editable false -listheight 300 -width 10

        set sorted_thickness_list \
		[ lsort -real -unique $IMCAGui::thickness_list ]

        foreach item $sorted_thickness_list {
                $grid.filter_desired insert list end $item
        }

#        set thickness [ get_17bm_filter_thickness ]

        set thickness 0

        if { $thickness == "Moving" } {
                set thickness 0
        }
        if { $thickness == "Fault" } {
                set thickness 0
        }

	$grid.filter_desired selection set 0

	button $grid.change -text "Change" -width 6 \
		-command [ itcl::code change_17bm_filters $record_list $grid ]

	$grid.change configure -background limegreen

	$grid.change configure -activebackground limegreen

	$grid.change configure -disabledforeground \
			[ $grid.change cget -foreground ]

        label $grid.filter_desired_units -text "um  "

        grid $grid.filter_label         -row 1 -column 0
        grid $grid.filter_current       -row 1 -column 1
        grid $grid.filter_current_units -row 1 -column 2 -sticky w
        grid $grid.filter_desired       -row 1 -column 3
        grid $grid.filter_desired_units -row 1 -column 4 -sticky w
        grid $grid.change               -row 1 -column 5

}

proc validate_17bm_filter_value { contents } {
	set valid [ iwidgets::Entryfield::integer $contents ]

	if { $valid == 0 } {
		return 0
	}
	if { $contents < 0 } {
		return 0
	}
	if { $contents > $IMCAGui::maximum_filter_thickness } {
		return 0
	}
	return 1
}

proc invalid_17bm_filter_value {} {
	set t "The filter thickness must be a value from 0 to "
	append t "$IMCAGui::maximum_filter_thickness um"

	tk_dialog .invalid_filter_dialog \
	"Invalid Input" $t error 0 "OK"
}

############################################################################

proc create_17bm_shutter_frame { record_list parent_frame } {

	set shutter_frame $parent_frame.shutter_frame

	set IMCAGui::shutter_frame $shutter_frame

	frame $shutter_frame
	pack $shutter_frame -side top

	label $shutter_frame.status -text "Status Unknown" -width 14 \
						-font $IMCAGui::shutter_font
	pack $shutter_frame.status -side top

	frame $shutter_frame.sep
	pack $shutter_frame.sep -side top -pady 5

	button $shutter_frame.button -text "Shutter\nButton" \
		-bg gray -activebackground gray \
		-command [ itcl::code change_17bm_shutter_status \
			$record_list $shutter_frame ]

	pack $shutter_frame.button -side top
}

############################################################################

proc create_17bm_feedback_frame { record_list parent_frame } {

	set feedback_frame $parent_frame.feedback_frame

	set IMCAGui::feedback_frame $feedback_frame

	frame $feedback_frame
	pack $feedback_frame -side top

	label $feedback_frame.title -text \
				"X-ray Beam Position\nFeedback Control"

	pack $feedback_frame.title -side top

	frame $feedback_frame.sep1
	pack $feedback_frame.sep1 -side top -pady 10

	button $feedback_frame.x_enable -text "X Feedback" \
	    -background gray -activebackground gray -command \
	    	[ itcl::code change_17bm_feedback $record_list "x_fb_enable" ]

	pack $feedback_frame.x_enable -side top

	frame $feedback_frame.sep2
	pack $feedback_frame.sep2 -side top -pady 10

	button $feedback_frame.y_enable -text "Y Feedback" \
	    -background gray -activebackground gray -command \
	    	[ itcl::code change_17bm_feedback $record_list "y_fb_enable" ]

	pack $feedback_frame.y_enable -side top
}

############################################################################

proc update_filter_frame { frame } {
	if { $IMCAGui::beamline_name == "imca17id" } {
		update_17id_filter_frame $frame
	} else {
		update_17bm_filter_frame $frame
	}
}

############################################################################
proc update_17id_filter_frame { frame } {
	global callback_values

       	set grid $frame.grid

       	set window $grid.filter_current
	set button $grid.move

       	set thickness [ get_17id_filter_thickness ]

         $window configure -text $thickness

       	if { $thickness == "Fault" } {
               	$window configure -fg red
       	} elseif { $thickness == "Moving" } {
               	$window configure -fg limegreen
       	} elseif { $thickness == "0.0" } {
               	$window configure -fg blue
       	} else {
               	$window configure -fg red
       	}

	set busy_list [ list ]
	lappend busy_list $callback_values(filter1_busy)
	lappend busy_list $callback_values(filter2_busy)

	compute_move_state $button $busy_list "Move"
}

############################################################################

proc update_17bm_filter_frame { frame } {
	global callback_values

       	set grid $frame.grid

       	set window $grid.filter_current
	set button $grid.move

       	set thickness [ get_17bm_filter_thickness ]

        $window configure -text $thickness

       	if { $thickness == "Fault" } {
               	$window configure -fg red
       	} elseif { $thickness == "Moving" } {
               	$window configure -fg limegreen
       	} elseif { $thickness == "0.0" } {
               	$window configure -fg blue
       	} else {
               	$window configure -fg red
       	}
}

############################################################################

proc update_17bm_shutter_frame { shutter_frame } {
	global callback_values

	set shutter_closed $callback_values(bshutter_closed)

	if { $shutter_closed == 0 } {
		$shutter_frame.status configure -text "B Shutter Open"
		$shutter_frame.status configure -bg limegreen

		$shutter_frame.button configure -text "Close\nShutter"
		$shutter_frame.button configure -bg limegreen
		$shutter_frame.button configure -activebackground limegreen
	} else {
		$shutter_frame.status configure -text "B Shutter Closed"
		$shutter_frame.status configure -bg red

		$shutter_frame.button configure -text "Open\nShutter"
		$shutter_frame.button configure -bg red
		$shutter_frame.button configure -activebackground red
	}
	$shutter_frame.button configure -state active
}

proc change_17bm_shutter_status { record_list shutter_frame } {
	global callback_values

	set shutter_open_record  [ $record_list get_record "bshutter_open" ]
	set shutter_close_record [ $record_list get_record "bshutter_close" ]

	set shutter_closed $callback_values(bshutter_closed)

	if { $shutter_closed == 0 } {
		$shutter_close_record send_variable 1
	} else {
		$shutter_open_record send_variable 1
	}

	$shutter_frame.button configure -bg yellow
	$shutter_frame.button configure -activebackground yellow
	$shutter_frame.button configure -state disabled

	update
}

############################################################################

proc update_17bm_feedback_frame { feedback_frame } {
	global callback_values

	set x_fb_enable $callback_values(x_fb_enable)

	set button $feedback_frame.x_enable

	if { $x_fb_enable } {
		$button configure -text "X Feedback\nEnabled"
		$button configure -bg limegreen -activebackground limegreen
	} else {
		$button configure -text "X Feedback\nDisabled"
		$button configure -bg red -activebackground red
	}

	set y_fb_enable $callback_values(y_fb_enable)

	set button $feedback_frame.y_enable

	if { $y_fb_enable } {
		$button configure -text "Y Feedback\nEnabled"
		$button configure -bg limegreen -activebackground limegreen
	} else {
		$button configure -text "Y Feedback\nDisabled"
		$button configure -bg red -activebackground red
	}
}

proc change_17bm_feedback { record_list record_name } {
	global callback_values

	set feedback_enable $callback_values($record_name)

	if { $feedback_enable } {
		set feedback_enable 0
	} else {
		set feedback_enable 1
	}

	set feedback_record [ $record_list get_record $record_name ]

	$feedback_record send_variable $feedback_enable
}

############################################################################

proc update_bpm_frame { frame } {
	if { $IMCAGui::beamline_name == "imca17id" } {
		update_17id_bpm_frame $frame
	} else {
		update_17bm_bpm_frame $frame
	}
}

############################################################################

proc update_17bm_bpm_frame { frame } {
	global callback_values

	$frame.bpm_plot element configure beamspot \
			-xdata $callback_values(bpm_x)

	$frame.bpm_plot element configure beamspot \
			-ydata $callback_values(bpm_y)

	$frame.xyframe.x_value configure -text $callback_values(bpm_x)

	$frame.xyframe.y_value configure -text $callback_values(bpm_y)
}

############################################################################

proc create_tab_optimize_window { record_list window } {

	global env

	set IMCAGui::optimize_window $window

#	if { $IMCAGui::beamline_name == "imca17id" } {
#		label $window.top_label -text "Sector 17-ID Optimization"
#	} else {
#		label $window.top_label -text "Sector 17-BM Optimization"
#	}
#	pack $window.top_label -side top
#
#	frame $window.sep1 -height 2 -borderwidth 1 -relief sunken
#	pack $window.sep1 -expand yes -fill x -pady 5

	label $window.auto_frame_label -text "Automatic Optimization"

	set auto_frame $window.auto_frame

	frame $auto_frame

	# Only show the Automatic Optimization stuff for imca17id.

	if { $IMCAGui::beamline_name == "imca17id" } {
		pack $window.auto_frame_label -side top
		pack $auto_frame -expand yes -fill x -pady 5
	}

	#----------------------------------------------------------

	set auto_mono_filename \
    [ file join $env(MXDIR) "etc" "optimize_auto" "optimize_auto_mono.tcl" ]

	set row 0

	label $auto_frame.mono_label \
		-text "Optimize Monochromator     "
	button $auto_frame.mono_button -text "Start" \
		-bg limegreen -activebackground limegreen \
		-command [ itcl::code auto_optimize \
			$record_list $window $auto_mono_filename]

	grid $auto_frame.mono_label -row $row -column 0 -sticky w
	grid $auto_frame.mono_button -row $row -column 1

	#----------------------------------------------------------

	if { $IMCAGui::beamline_name == "imca17id" } {
		set auto_slits_filename \
    [ file join $env(MXDIR) "etc" "optimize_auto" "optimize_auto_slits.tcl" ]

		incr row

		label $auto_frame.slits_label \
			-text "Optimize Slits"
		button $auto_frame.slits_button -text "Start" \
			-bg limegreen -activebackground limegreen \
			-command [ itcl::code auto_optimize \
				$record_list $window $auto_slits_filename]

		grid $auto_frame.slits_label -row $row -column 0 -sticky w
		grid $auto_frame.slits_button -row $row -column 1

		#----------------------------------------------------------

		set auto_table_filename \
    [ file join $env(MXDIR) "etc" "optimize_auto" "optimize_auto_table.tcl" ]

		incr row

		label $auto_frame.table_label \
			-text "Optimize Table"
		button $auto_frame.table_button -text "Start" \
			-bg limegreen -activebackground limegreen \
			-command [ itcl::code auto_optimize \
				$record_list $window $auto_table_filename]

		grid $auto_frame.table_label -row $row -column 0 -sticky w
		grid $auto_frame.table_button -row $row -column 1
	}

	#----------------------------------------------------------

	if { $IMCAGui::beamline_name == "imca17bm" } {
		set auto_mirror_filename \
    [ file join $env(MXDIR) "etc" "optimize_auto" "optimize_auto_mirror.tcl" ]

		incr row

		label $auto_frame.mirror_label \
			-text "Optimize Mirror"
		button $auto_frame.mirror_button -text "Start" \
			-bg limegreen -activebackground limegreen \
			-command [ itcl::code auto_optimize \
				$record_list $window $auto_mirror_filename]

		grid $auto_frame.mirror_label -row $row -column 0 -sticky w
		grid $auto_frame.mirror_button -row $row -column 1
	}

	#----------------------------------------------------------

	set auto_all_filename \
    [ file join $env(MXDIR) "etc" "optimize_auto" "optimize_auto_all.tcl" ]

	incr row

	label $auto_frame.all_label \
		-text "Optimize All"
	button $auto_frame.all_button -text "Start" \
		-bg limegreen -activebackground limegreen \
		-command [ itcl::code auto_optimize \
			$record_list $window $auto_all_filename]

	grid $auto_frame.all_label -row $row -column 0 -sticky w
	grid $auto_frame.all_button -row $row -column 1

	#==========================================================

	if { $IMCAGui::beamline_name == "imca17id" } {
		frame $window.sep2 -height 2 -borderwidth 1 -relief sunken
		pack $window.sep2 -expand yes -fill x -pady 5
	}

	label $window.manual_frame_label -text "Manual Optimization"
	pack $window.manual_frame_label -side top

	set manual_frame $window.manual_frame

	frame $manual_frame
	pack $manual_frame -expand yes -fill x -pady 5 -padx 10

	set row 0
	set statuswidth 30

	#----------------------------------------------------------

	label $manual_frame.mono_label -text Monochromator
	label $manual_frame.status_header -text Status
	label $manual_frame.old_header -text "Old Position"
	label $manual_frame.new_header -text "New Position"

	grid $manual_frame.mono_label		-row $row -column 0 -sticky w
	grid $manual_frame.status_header	-row $row -column 2
	grid $manual_frame.old_header		-row $row -column 3
	grid $manual_frame.new_header		-row $row -column 4

	#----------------------------------------------------------

	incr row

	set motor_name "omega"

	set omega [ $record_list get_record $motor_name ]

	set thisfield $manual_frame.omega

	set IMCAGui::motor_name_to_field_name($motor_name) $thisfield
	
	button ${thisfield}_button -text "Optimize Omega" \
		-bg limegreen -activebackground limegreen -command \
		[ itcl::code optimize_omega $record_list $omega $thisfield ]
	
	label ${thisfield}_status -width $statuswidth -text "Not Yet Optimized"
	
	label ${thisfield}_old_position -text [ optimize_get_position $omega ]
	label ${thisfield}_new_position -text "---"
	
	grid ${thisfield}_button       -row $row -column 0 -sticky w
	grid ${thisfield}_status       -row $row -column 2 -sticky w
	grid ${thisfield}_old_position -row $row -column 3 -sticky e
	grid ${thisfield}_new_position -row $row -column 4 -sticky e
	
	#-----------------------------------------------------------
	
	if { $IMCAGui::beamline_name == "imca17id" } {
		# Code only for APS Sector 17-ID

		incr row

		set motor_name "chi"

		set chi [ $record_list get_record $motor_name ]

		set thisfield $manual_frame.chi
	
		set IMCAGui::motor_name_to_field_name($motor_name) $thisfield
	
		button ${thisfield}_button -text "Optimize Chi" \
			-bg limegreen -activebackground limegreen -command \
			[ itcl::code optimize_chi $record_list $chi $thisfield ]
	
		label ${thisfield}_status -width $statuswidth -text "Not Yet Optimized"
	
		label ${thisfield}_old_position -text [ optimize_get_position $chi ]
		label $manual_frame.chi_new_position -text "---"
	
		grid ${thisfield}_button       -row $row -column 0 -sticky w
		grid ${thisfield}_status       -row $row -column 2 -sticky w
		grid ${thisfield}_old_position -row $row -column 3 -sticky e
		grid ${thisfield}_new_position -row $row -column 4 -sticky e
	
	#----------------------------------------------------------

#	if { $IMCAGui::beamline_name == "imca17bm" } {
#
#		# Code only for APS Sector 17-BM
#
#		label $manual_frame.blankx
#		grid $manual_frame.blankx
#	
#		incr row
#
#		set motor_name "focus"
#	
#		set focus [ $record_list get_record $motor_name ]
#
#		set thisfield $manual_frame.focus
#
#		set IMCAGui::motor_name_to_field_name($motor_name) $thisfield
#	
#		button ${thisfield}_button -text "Optimize Focus" \
#		  -background limegreen -activebackground limegreen -command \
#	[ itcl::code optimize_focus $record_list $focus $thisfield ]
#
#		label ${thisfield}_status -width $statuswidth -text \
#			"Not Yet Optimized"
#	
#		label ${thisfield}_old_position -text \
#			[ optimize_get_position $focus ]
#
#		label ${thisfield}_new_position -text "---"
#
#		grid ${thisfield}_button       -row $row -column 0 -sticky w
#		grid ${thisfield}_status       -row $row -column 2 -sticky w
#		grid ${thisfield}_old_position -row $row -column 3 -sticky e
#		grid ${thisfield}_new_position -row $row -column 4 -sticky e
#	
#		incr row
#
#		# End of code only for APS Sector 17-BM
#	}

		#-----------------------------------------------------------
	
		label $manual_frame.blank1
		grid $manual_frame.blank1
	
		incr row
	
		#-----------------------------------------------------------
	
		label $manual_frame.mirror -text Mirror
		grid $manual_frame.mirror -sticky w
	
		incr row

		#-----------------------------------------------------------
	
		if { $IMCAGui::beamline_name == "imca17id" } {
			set motor_name "vmirror_z"
		} else {
			set motor_name "mjack_vert"
		}

		set vert_mirror [ $record_list get_record $motor_name ]

		set thisfield $manual_frame.vert_mirror
	
		set IMCAGui::motor_name_to_field_name($motor_name) $thisfield
	
		button ${thisfield}_button -text "Optimize Vertical" \
	    -background limegreen -activebackground limegreen -command \
    [ itcl::code optimize_vert_mirror $record_list $vert_mirror $thisfield ]
	
		label ${thisfield}_status -width $statuswidth -text "Not Yet Optimized"
	
		label ${thisfield}_old_position -text \
			[ optimize_get_position $vert_mirror ]

		label ${thisfield}_new_position -text "---"
	
		grid ${thisfield}_button       -row $row -column 0 -sticky w
		grid ${thisfield}_status       -row $row -column 2 -sticky w
		grid ${thisfield}_old_position -row $row -column 3 -sticky e
		grid ${thisfield}_new_position -row $row -column 4 -sticky e

		if { $IMCAGui::beamline_name == "imca17id" } {
			${thisfield}_button configure -state disabled
		}
	
		incr row
	
		# WML - Old start of imca17id only code.

		label $manual_frame.blank2
		grid $manual_frame.blank2
	
		incr row
	
		#-----------------------------------------------------------
	
		label $manual_frame.slits -text Slits
		grid $manual_frame.slits -sticky w
	
		incr row

		#-----------------------------------------------------------

		set motor_name "bv_center"
	
		set vert_slit [ $record_list get_record $motor_name ]

		set thisfield $manual_frame.vert_slit

		set IMCAGui::motor_name_to_field_name($motor_name) $thisfield
	
		button ${thisfield}_button -text "Optimize Vertical" \
	    -background limegreen -activebackground limegreen -command \
	    [ itcl::code optimize_vert_slit $record_list $vert_slit $thisfield ]
	
		label ${thisfield}_status -width $statuswidth \
					-text "Not Yet Optimized"
	
		label ${thisfield}_old_position -text \
			[ optimize_get_position $vert_slit ]

		label ${thisfield}_new_position -text "---"
	
		grid ${thisfield}_button       -row $row -column 0 -sticky w
		grid ${thisfield}_status       -row $row -column 2 -sticky w
		grid ${thisfield}_old_position -row $row -column 3 -sticky e
		grid ${thisfield}_new_position -row $row -column 4 -sticky e
	
		incr row
	
		#-----------------------------------------------------------

		set motor_name "bh_center"
	
		set horiz_slit [ $record_list get_record $motor_name ]

		set thisfield $manual_frame.horiz_slit
	
		set IMCAGui::motor_name_to_field_name($motor_name) $thisfield
	
		button ${thisfield}_button -text "Optimize Horizontal" \
	    -background limegreen -activebackground limegreen -command \
    [ itcl::code optimize_horiz_slit $record_list $horiz_slit $thisfield ]
	
		label ${thisfield}_status -width $statuswidth \
			-text "Not Yet Optimized"
	
		label ${thisfield}_old_position -text \
			[ optimize_get_position $horiz_slit ]

		label ${thisfield}_new_position -text "---"
	
		grid ${thisfield}_button       -row $row -column 0 -sticky w
		grid ${thisfield}_status       -row $row -column 2 -sticky w
		grid ${thisfield}_old_position -row $row -column 3 -sticky e
		grid ${thisfield}_new_position -row $row -column 4 -sticky e
	
		incr row
	
		#-----------------------------------------------------------
	
		label $manual_frame.blank3
		grid $manual_frame.blank3
	
		incr row
	
		label $manual_frame.table -text Table
		grid $manual_frame.table          -row $row -column 0 -sticky w

		label $manual_frame.halfslit_label -text "Half Slits"
		grid $manual_frame.halfslit_label -row $row -column 1
	
		incr row

		#-----------------------------------------------------------

		set motor_name "tz"
	
		set table_vert [ $record_list get_record $motor_name ]

		set thisfield $manual_frame.table_vert

		set IMCAGui::motor_name_to_field_name($motor_name) $thisfield
	
		button ${thisfield}_button -text "Optimize Vertical" \
	    -background limegreen -activebackground limegreen -command \
	[ itcl::code optimize_table_vert $record_list $table_vert $thisfield ]

		label ${thisfield}_status -width $statuswidth -text \
			"Not Yet Optimized"
	
		label ${thisfield}_old_position -text \
			[ optimize_get_position $table_vert ]

		label ${thisfield}_new_position -text "---"

		set shutter_frame \
		    [ create_shutter_frame $thisfield $record_list "us_vert" ]
	
		grid ${thisfield}_button       -row $row -column 0 -sticky w
		grid $shutter_frame            -row $row -column 1 -sticky w
		grid ${thisfield}_status       -row $row -column 2 -sticky w
		grid ${thisfield}_old_position -row $row -column 3 -sticky e
		grid ${thisfield}_new_position -row $row -column 4 -sticky e
	
		incr row

		#-----------------------------------------------------------

		set motor_name "ty"
	
		set table_horiz [ $record_list get_record $motor_name ]

		set thisfield $manual_frame.table_horiz

		set IMCAGui::motor_name_to_field_name($motor_name) $thisfield
	
		button ${thisfield}_button -text "Optimize Horizontal" \
	    -background limegreen -activebackground limegreen -command \
	[ itcl::code optimize_table_horiz $record_list $table_horiz $thisfield ]

		label ${thisfield}_status -width $statuswidth -text \
			"Not Yet Optimized"
	
		label ${thisfield}_old_position -text \
			[ optimize_get_position $table_horiz ]

		label ${thisfield}_new_position -text "---"
	
		set shutter_frame \
		    [ create_shutter_frame $thisfield $record_list "us_horiz" ]
	
		grid ${thisfield}_button       -row $row -column 0 -sticky w
		grid $shutter_frame            -row $row -column 1 -sticky w
		grid ${thisfield}_status       -row $row -column 2 -sticky w
		grid ${thisfield}_old_position -row $row -column 3 -sticky e
		grid ${thisfield}_new_position -row $row -column 4 -sticky e
	
		incr row

		#-----------------------------------------------------------

		set motor_name "tpitch"
	
		set table_pitch [ $record_list get_record $motor_name ]

		set thisfield $manual_frame.table_pitch

		set IMCAGui::motor_name_to_field_name($motor_name) $thisfield
	
		button ${thisfield}_button -text "Optimize Pitch" \
	    -background limegreen -activebackground limegreen -command \
    [ itcl::code optimize_table_pitch $record_list $table_pitch $thisfield ]

		label ${thisfield}_status -width $statuswidth -text \
			"Not Yet Optimized"
	
		label ${thisfield}_old_position -text \
			[ optimize_get_position $table_pitch ]

		label ${thisfield}_new_position -text "---"
	
		set shutter_frame \
		    [ create_shutter_frame $thisfield $record_list "ds_vert" ]
	
		grid ${thisfield}_button       -row $row -column 0 -sticky w
		grid $shutter_frame            -row $row -column 1 -sticky w
		grid ${thisfield}_status       -row $row -column 2 -sticky w
		grid ${thisfield}_old_position -row $row -column 3 -sticky e
		grid ${thisfield}_new_position -row $row -column 4 -sticky e
	
		incr row

		#-----------------------------------------------------------

		set motor_name "tyaw"
	
		set table_yaw [ $record_list get_record $motor_name ]

		set thisfield $manual_frame.table_yaw

		set IMCAGui::motor_name_to_field_name($motor_name) $thisfield
	
		button ${thisfield}_button -text "Optimize Yaw" \
	    -background limegreen -activebackground limegreen -command \
	[ itcl::code optimize_table_yaw $record_list $table_yaw $thisfield ]

		label ${thisfield}_status -width $statuswidth -text \
			"Not Yet Optimized"
	
		label ${thisfield}_old_position -text \
			[ optimize_get_position $table_yaw ]

		label ${thisfield}_new_position -text "---"
	
		set shutter_frame \
		    [ create_shutter_frame $thisfield $record_list "ds_horiz" ]
	
		grid ${thisfield}_button       -row $row -column 0 -sticky w
		grid $shutter_frame            -row $row -column 1 -sticky w
		grid ${thisfield}_status       -row $row -column 2 -sticky w
		grid ${thisfield}_old_position -row $row -column 3 -sticky e
		grid ${thisfield}_new_position -row $row -column 4 -sticky e

		incr row
	
		#-----------------------------------------------------------
	
		label $manual_frame.spacer -text ""
		grid $manual_frame.spacer          -row $row -column 0 -sticky w

		incr row

		#-----------------------------------------------------------
	
		button $manual_frame.table_control_start \
				-text "Adjust Table Motors" \
	    			-background limegreen \
				-activebackground limegreen -command \
				[ itcl::code eval exec "adsc_table &" ]

		grid $manual_frame.table_control_start -row $row -column 0
	
		# End of code only for APS Sector 17-ID
	}

	#-----------------------------------------------------------

}

proc disable_button { button_name } {
	catch { $button_name configure -state disabled }
}

proc enable_button { button_name } {
	catch { $button_name configure -state normal }
}

proc disable_buttons {} {
	set auto_frame ${IMCAGui::optimize_window}.auto_frame

	disable_button $auto_frame.mono_button
	disable_button $auto_frame.slits_button
	disable_button $auto_frame.table_button
	disable_button $auto_frame.mirror_button
	disable_button $auto_frame.all_button

	set manual_frame ${IMCAGui::optimize_window}.manual_frame

	disable_button $manual_frame.omega_button
	disable_button $manual_frame.chi_button

	if { $IMCAGui::beamline_name == "imca17bm" } {
		disable_button $manual_frame.focus_button
	}

	disable_button $manual_frame.vert_mirror_button
	disable_button $manual_frame.vert_slit_button
	disable_button $manual_frame.horiz_slit_button

	if { $IMCAGui::beamline_name == "imca17id" } {
		disable_button $manual_frame.table_vert_button
		disable_button $manual_frame.table_horiz_button
		disable_button $manual_frame.table_pitch_button
		disable_button $manual_frame.table_yaw_button
	}

	if { $IMCAGui::beamline_name == "imca17bm" } {
		set staff_frame ${IMCAGui::staff_optimize_window}.manual_frame

		disable_button $staff_frame.omega_button
		disable_button $staff_frame.chi_button
		disable_button $staff_frame.focus_button
		disable_button $staff_frame.vert_mirror_button
	}
}

proc enable_buttons {} {
	set auto_frame ${IMCAGui::optimize_window}.auto_frame

	enable_button $auto_frame.mono_button
	enable_button $auto_frame.slits_button
	enable_button $auto_frame.table_button
	enable_button $auto_frame.mirror_button
	enable_button $auto_frame.all_button

	set manual_frame ${IMCAGui::optimize_window}.manual_frame

	enable_button $manual_frame.omega_button
	enable_button $manual_frame.chi_button

	if { $IMCAGui::beamline_name == "imca17bm" } {
		enable_button $manual_frame.focus_button
		enable_button $manual_frame.vert_mirror_button
	}

	enable_button $manual_frame.vert_slit_button
	enable_button $manual_frame.horiz_slit_button

	if { $IMCAGui::beamline_name == "imca17id" } {
		enable_button $manual_frame.table_vert_button
		enable_button $manual_frame.table_horiz_button
		enable_button $manual_frame.table_pitch_button
		enable_button $manual_frame.table_yaw_button
	}

	if { $IMCAGui::beamline_name == "imca17bm" } {
		set staff_frame ${IMCAGui::staff_optimize_window}.manual_frame

		enable_button $staff_frame.omega_button
		enable_button $staff_frame.chi_button
#		enable_button $staff_frame.focus_button
		enable_button $staff_frame.vert_mirror_button
	}
}

proc optimize_omega { record_list motor guifield } {
	perform_optimization $record_list $motor $guifield "fit_to_peak"
}

proc optimize_chi { record_list motor guifield } {
	perform_optimization $record_list $motor $guifield "fit_to_peak"
}

proc optimize_focus { record_list motor guifield } {

	perform_optimization $record_list $motor $guifield "fit_to_peak"
}

proc optimize_horiz_slit { record_list motor guifield } {

	perform_optimization $record_list $motor $guifield "fit_to_peak"
}

proc optimize_vert_slit { record_list motor guifield } {

	perform_optimization $record_list $motor $guifield "fit_to_peak"
}

proc optimize_vert_mirror { record_list motor guifield } {

	perform_optimization $record_list $motor $guifield "fit_to_peak"
}

proc optimize_table_vert { record_list motor guifield } {

	perform_optimization $record_list $motor $guifield "fit_to_step"
}

proc optimize_table_horiz { record_list motor guifield } {

	perform_optimization $record_list $motor $guifield "fit_to_step"
}

proc optimize_table_pitch { record_list motor guifield } {

	perform_optimization $record_list $motor $guifield "fit_to_step"
}

proc optimize_table_yaw { record_list motor guifield } {

	perform_optimization $record_list $motor $guifield "fit_to_step"
}

############################################################################

proc create_17bm_staff_optimize_window { record_list window } {

	global env

	set IMCAGui::staff_optimize_window $window

	#==========================================================

	label $window.manual_frame_label -text "Manual Optimization"
	pack $window.manual_frame_label -side top

	set manual_frame $window.manual_frame

	frame $manual_frame
	pack $manual_frame -expand yes -fill x -pady 5 -padx 10

	set row 0
	set statuswidth 30

	#----------------------------------------------------------

	label $manual_frame.mono_label -text Monochromator
	label $manual_frame.status_header -text Status
	label $manual_frame.old_header -text "Old Position"
	label $manual_frame.new_header -text "New Position"

	grid $manual_frame.mono_label		-row $row -column 0 -sticky w
	grid $manual_frame.status_header	-row $row -column 2
	grid $manual_frame.old_header		-row $row -column 3
	grid $manual_frame.new_header		-row $row -column 4

	#----------------------------------------------------------

	incr row

	set motor_name "omega"

	set omega [ $record_list get_record $motor_name ]

	set thisfield $manual_frame.omega

	set IMCAGui::motor_name_to_field_name($motor_name) $thisfield
	
	button ${thisfield}_button -text "Optimize Omega" \
		-bg limegreen -activebackground limegreen -command \
		[ itcl::code optimize_omega $record_list $omega $thisfield ]
	
	label ${thisfield}_status -width $statuswidth -text "Not Yet Optimized"
	
	label ${thisfield}_old_position -text [ optimize_get_position $omega ]
	label ${thisfield}_new_position -text "---"
	
	grid ${thisfield}_button       -row $row -column 0 -sticky w
	grid ${thisfield}_status       -row $row -column 2 -sticky w
	grid ${thisfield}_old_position -row $row -column 3 -sticky e
	grid ${thisfield}_new_position -row $row -column 4 -sticky e
	
	#----------------------------------------------------------

	incr row

	set motor_name "chi"

	set chi [ $record_list get_record $motor_name ]

	set thisfield $manual_frame.chi
	
	set IMCAGui::motor_name_to_field_name($motor_name) $thisfield
	
	button ${thisfield}_button -text "Optimize Chi" \
		-bg limegreen -activebackground limegreen -command \
		[ itcl::code optimize_chi $record_list $chi $thisfield ]
	
	label ${thisfield}_status -width $statuswidth -text "Not Yet Optimized"
	
	label ${thisfield}_old_position -text [ optimize_get_position $chi ]
	label $manual_frame.chi_new_position -text "---"
	
	grid ${thisfield}_button       -row $row -column 0 -sticky w
	grid ${thisfield}_status       -row $row -column 2 -sticky w
	grid ${thisfield}_old_position -row $row -column 3 -sticky e
	grid ${thisfield}_new_position -row $row -column 4 -sticky e
	
	#----------------------------------------------------------

	label $manual_frame.blankx
	grid $manual_frame.blankx
	
	incr row

	set motor_name "focus"
	
	set focus [ $record_list get_record $motor_name ]

	set thisfield $manual_frame.focus

	set IMCAGui::motor_name_to_field_name($motor_name) $thisfield
	
	button ${thisfield}_button -text "Optimize Focus" \
		  -background limegreen -activebackground limegreen -command \
	[ itcl::code optimize_focus $record_list $focus $thisfield ] \
		  -state disabled

	label ${thisfield}_status -width $statuswidth -text \
			"Not Yet Optimized"
	
	label ${thisfield}_old_position -text \
			[ optimize_get_position $focus ]

	label ${thisfield}_new_position -text "---"

	grid ${thisfield}_button       -row $row -column 0 -sticky w
	grid ${thisfield}_status       -row $row -column 2 -sticky w
	grid ${thisfield}_old_position -row $row -column 3 -sticky e
	grid ${thisfield}_new_position -row $row -column 4 -sticky e
	
	incr row

	#-----------------------------------------------------------
	
	label $manual_frame.blank1
	grid $manual_frame.blank1
	
	incr row
	
	#-----------------------------------------------------------
	
	label $manual_frame.mirror -text Mirror
	grid $manual_frame.mirror -sticky w
	
	incr row

	#-----------------------------------------------------------
	
	if { $IMCAGui::beamline_name == "imca17id" } {
		set motor_name "vmirror_z"
	} else {
		set motor_name "mjack_vert"
	}

	set vert_mirror [ $record_list get_record $motor_name ]

	set thisfield $manual_frame.vert_mirror
	
	set IMCAGui::motor_name_to_field_name($motor_name) $thisfield
	
	button ${thisfield}_button -text "Optimize Vertical" \
	    -background limegreen -activebackground limegreen -command \
    [ itcl::code optimize_vert_mirror $record_list $vert_mirror $thisfield ]
	
	label ${thisfield}_status -width $statuswidth -text "Not Yet Optimized"
	
	label ${thisfield}_old_position -text \
		[ optimize_get_position $vert_mirror ]

	label ${thisfield}_new_position -text "---"
	
	grid ${thisfield}_button       -row $row -column 0 -sticky w
	grid ${thisfield}_status       -row $row -column 2 -sticky w
	grid ${thisfield}_old_position -row $row -column 3 -sticky e
	grid ${thisfield}_new_position -row $row -column 4 -sticky e

	if { $IMCAGui::beamline_name == "imca17id" } {
		${thisfield}_button configure -state disabled
	}
	
	incr row
	
	#==========================================================

	frame $window.sep2 -height 2 -borderwidth 1 -relief sunken
	pack $window.sep2 -expand yes -fill x -pady 5

	label $window.tweak_frame_label -text "Tweak"
	pack $window.tweak_frame_label -side top

	set tweak_frame $window.tweak_frame

	frame $tweak_frame
	pack $tweak_frame -expand yes -fill x -pady 5 -padx 10

	set row 0

	#-----------------------------------------------------------

	set thisfield $tweak_frame.focus

	create_arrows

	set focus_record [ $record_list get_record "focus" ]

	set focus_position [ $focus_record get_position ]

	label ${thisfield}_name -text "Focus:"
	label ${thisfield}_value -fg blue -text "---"

	button ${thisfield}_move_negative -image left_arrow -command \
		[ itcl::code relative_focus_move $focus_record $thisfield -5 ]

	label ${thisfield}_setpoint -text $focus_position

	label ${thisfield}_hidden_left_limit -text \
		[ expr ( $focus_position - 50.0 ) ]

	label ${thisfield}_hidden_right_limit -text \
		[ expr ( $focus_position + 50.0 ) ]

	button ${thisfield}_move_positive -image right_arrow -command \
		[ itcl::code relative_focus_move $focus_record $thisfield  5 ]

	grid ${thisfield}_name          -row $row -column 0 -sticky w
	grid ${thisfield}_value         -row $row -column 1 -sticky w -padx 5
	grid ${thisfield}_move_negative -row $row -column 2
	grid ${thisfield}_setpoint      -row $row -column 3 -sticky w -padx 5
	grid ${thisfield}_move_positive -row $row -column 4

	add_imcagui_callback staff_optimize_callback
}

proc relative_focus_move { focus_record thisfield relative_distance } {

	global callback_values

	set left_limit [ ${thisfield}_hidden_left_limit cget -text ]
	set right_limit [ ${thisfield}_hidden_right_limit cget -text ]

	set current_position $callback_values(focus)

	set new_setpoint [ expr ( $current_position + $relative_distance ) ]

	if { $new_setpoint < $left_limit } {
		set new_setpoint $left_limit

		log_puts "focus tweak limit at $left_limit reached."
	}
	if { $new_setpoint > $right_limit } {
		set new_setpoint $right_limit

		log_puts "focus tweak limit at $right_limit reached."
	}

	${thisfield}_setpoint configure -text $new_setpoint

	$focus_record move_absolute $new_setpoint
}

proc create_arrows {} {

	image create bitmap left_arrow -foreground limegreen -data "
#define imcagui_left_width 16
#define imcagui_left_height 16
static unsigned char imcagui_left_bits[] = {
   0x00, 0xc0, 0x00, 0xf0, 0x00, 0xfc, 0x00, 0xff, 0xc0, 0xff, 0xf0, 0xff,
   0xfc, 0xff, 0xff, 0xff, 0xff, 0xff, 0xfc, 0xff, 0xf0, 0xff, 0xc0, 0xff,
   0x00, 0xff, 0x00, 0xfc, 0x00, 0xf0, 0x00, 0xc0};
"

	image create bitmap right_arrow -foreground limegreen -data "
#define imcagui_right_width 16
#define imcagui_right_height 16
static unsigned char imcagui_right_bits[] = {
   0x03, 0x00, 0x0f, 0x00, 0x3f, 0x00, 0xff, 0x00, 0xff, 0x03, 0xff, 0x0f,
   0xff, 0x3f, 0xff, 0xff, 0xff, 0xff, 0xff, 0x3f, 0xff, 0x0f, 0xff, 0x03,
   0xff, 0x00, 0x3f, 0x00, 0x0f, 0x00, 0x03, 0x00};
"

}

proc staff_optimize_callback { } {

	global callback_values

	set window ${IMCAGui::staff_optimize_window}.tweak_frame

	set focus $callback_values(focus)

	set focus_value_window $window.focus_value

	$focus_value_window configure -text $focus
}

############################################################################

proc create_shutter_frame { thisfield record_list shutter_name } {

	set shutter_record [ $record_list get_record $shutter_name ]

	set status ${shutter_name}_status

	set frame ${thisfield}_shutter_frame

	frame $frame

	label $frame.pad -text "   "
	pack $frame.pad -side left

	radiobutton $frame.removed -text "Removed" -variable $status -value 0

	pack $frame.removed -side left

	radiobutton $frame.inserted -text "Inserted" -variable $status -value 1

	pack $frame.inserted -side left

	set relay_status [ $shutter_record get_relay_status ]

	if { $relay_status == 0 } {
		$frame.removed select
	} else {
		$frame.inserted select
	}

	$frame.removed configure -command \
			[ itcl::code $shutter_record relay_command 0 ]

	$frame.inserted configure -command \
			[ itcl::code $shutter_record relay_command 1 ]

	return $frame
}

############################################################################

proc auto_optimize { record_list window filename } {

	log_puts "Running optimization script '$filename'"

	set manual_frame $window.manual_frame

	set status [ catch { source $filename } result ]

	if { $status != 0 } {
		auto_optimize_failed $manual_frame $filename $result
	}

	disable_buttons

	$record_list set_plot_enable 2

#	set status [ catch { main $record_list } result ]

	set status 0
	main $record_list

	$record_list set_plot_enable 1

	enable_buttons

	if { $status != 0 } {
		auto_optimize_failed $manual_frame $filename $result
	}
}

proc auto_optimize_failed { manual_frame filename result } {

	set status_header $manual_frame.status_header

	set status_list [ grid slaves $manual_frame -column 2 ]

	foreach win $status_list {
		log_puts "win = $win"

		if { $win != $status_header } {
			$win configure -fg red -text "Auto optimization failed"
		}
	}

	set t "The automatic optimization script '$filename' failed.  "
	append t "Error = '$result'"

	log_puts $t

	Mx::error_message "Automatic Optimization Failed." $t "" ""
}

############################################################################

proc create_staff_window { record_list window } {

	set IMCAGui::staff_window $window

#	frame $window.sep1 -height 2 -borderwidth 1 -relief flat
#	pack $window.sep1 -expand yes -fill x -pady 25 -side top

	set grid $window.grid

	frame $grid
	pack $grid -side top

	set row 0

	label $grid.edge_label -text "Edge Energy: " -width 25
	label $grid.edge_value -text "-----" -fg blue -width 10

	iwidgets::entryfield $grid.edge_desired -width 12

	$grid.edge_desired clear

	$grid.edge_desired configure \
    -command [ itcl::code change_edge_energy $record_list $grid.edge_desired ]

	label $grid.edge_units -text eV

	grid $grid.edge_label   -row $row -column 0 -padx 2 -sticky w
	grid $grid.edge_value   -row $row -column 1 -padx 2 -sticky w
	grid $grid.edge_desired -row $row -column 2 -padx 2 -sticky w
	grid $grid.edge_units   -row $row -column 3 -padx 2 -sticky w

	#---------------------------------------------------------------

	if { $IMCAGui::beamline_name == "imca17id" } {
		incr row

		label $grid.id_ev_enabled_label \
				-text "ID_ev_enabled: " -width 25
		label $grid.id_ev_enabled_value \
				-text "-----" -fg blue -width 10

		iwidgets::entryfield $grid.id_ev_enabled_desired -width 12

		$grid.id_ev_enabled_desired clear

		$grid.id_ev_enabled_desired configure -command [ itcl::code \
		change_id_ev_enabled $record_list $grid.id_ev_enabled_desired ]

		grid $grid.id_ev_enabled_label \
					-row $row -column 0 -padx 2 -sticky w
		grid $grid.id_ev_enabled_value \
					-row $row -column 1 -padx 2 -sticky w
		grid $grid.id_ev_enabled_desired \
					-row $row -column 2 -padx 2 -sticky w

		#---------------------------------------------------------------

		incr row

		label $grid.id_ev_harmonic_label \
				-text "ID_ev_harmonic: " -width 25
		label $grid.id_ev_harmonic_value \
				-text "-----" -fg blue -width 10

		iwidgets::entryfield $grid.id_ev_harmonic_desired -width 12

		$grid.id_ev_harmonic_desired clear

		$grid.id_ev_harmonic_desired configure -command [ itcl::code \
	    change_id_ev_harmonic $record_list $grid.id_ev_harmonic_desired ]

		grid $grid.id_ev_harmonic_label \
					-row $row -column 0 -padx 2 -sticky w
		grid $grid.id_ev_harmonic_value \
					-row $row -column 1 -padx 2 -sticky w
		grid $grid.id_ev_harmonic_desired \
					-row $row -column 2 -padx 2 -sticky w

		#---------------------------------------------------------------

		incr row

		label $grid.id_ev_params_label -text "ID_ev_params: " -width 25
		label $grid.id_ev_params_value -text "-----" -fg blue -width 10

		iwidgets::entryfield $grid.id_ev_params_desired -width 12

		$grid.id_ev_params_desired clear

		$grid.id_ev_params_desired configure -command [ itcl::code \
	    change_id_ev_params $record_list $grid.id_ev_params_desired ]

		grid $grid.id_ev_params_label \
					-row $row -column 0 -padx 2 -sticky w
		grid $grid.id_ev_params_value \
					-row $row -column 1 -padx 2 -sticky w
		grid $grid.id_ev_params_desired \
					-row $row -column 2 -padx 2 -sticky w
	}

	#---------------------------------------------------------------

	add_imcagui_callback staff_window_callback

	#---------------------------------------------------------------

	frame $window.sep1 -height 2 -borderwidth 1 -relief flat
	pack $window.sep1 -fill x -pady 10

	set button_grid $window.button_grid

	frame $button_grid
	pack $button_grid -side top

	set column 0

	#---------------------------------------------------------------

	button $button_grid.mxgui_button \
		-bg limegreen -activebackground limegreen \
		-text "Start mxgui" \
		-command [ itcl::code eval exec "mxgui &" ]

	grid $button_grid.mxgui_button -row 0 -column $column \
		-padx 5 -pady 5 -sticky nsew

	incr column

	#---------------------------------------------------------------

	if { $IMCAGui::beamline_name == "imca17id" } {
		button $button_grid.medm_button \
			-bg limegreen -activebackground limegreen \
			-text "Start s17id" \
			-command [ itcl::code eval exec "s17id &" ]
	} else {
		button $button_grid.medm_button \
			-bg limegreen -activebackground limegreen \
			-text "Start s17bm" \
			-command [ itcl::code eval exec "s17bm &" ]
	}

	grid $button_grid.medm_button -row 0 -column $column \
		-padx 5 -pady 5 -sticky nsew

	incr column

	#---------------------------------------------------------------

	button $button_grid.motor_button \
		-bg limegreen -activebackground limegreen \
		-text "Start motor" \
		-command [ itcl::code eval exec "xterm -sb -e 'motor_xterm' &" ]

	grid $button_grid.motor_button -row 0 -column $column \
		-padx 5 -pady 5 -sticky nsew

	incr column

	set column 0

	#---------------------------------------------------------------

	button $button_grid.status_button \
		-bg limegreen -activebackground limegreen \
		-text "Ring Status" \
		-command [ itcl::code eval exec "medm -x SR_Status.adl &" ]

	grid $button_grid.status_button -row 1 -column $column \
		-padx 5 -pady 5 -sticky nsew

	incr column

	#---------------------------------------------------------------

	if { $IMCAGui::beamline_name == "imca17id" } {
		set command \
	"medm -x -macro \"sector=17,sector0=17,sectorPlusOne=18\" IDxbpm.adl &"

		button $button_grid.steering_button \
			-bg limegreen -activebackground limegreen \
			-text "Steering Status" \
			-command [ itcl::code eval exec $command  ]
	} else {
		set command \
			"medm -x -macro \"sector=17,sector0=17\" BMxbpm.adl &"

		button $button_grid.steering_button \
			-bg limegreen -activebackground limegreen \
			-text "Steering Status" \
			-command [ itcl::code eval exec $command ]
	}

	grid $button_grid.steering_button -row 1 -column $column \
		-padx 5 -pady 5 -sticky nsew

	incr column

	#---------------------------------------------------------------

	if { $IMCAGui::beamline_name == "imca17id" } {
		set command "medm -x -macro \"xx=17\" UndulatorCtl.adl &"

		button $button_grid.id_button \
			-bg limegreen -activebackground limegreen \
			-text "Undulator Control" \
			-command [ itcl::code eval exec $command  ]

		grid $button_grid.id_button -row 1 -column $column \
			-padx 5 -pady 5 -sticky nsew
	}

	#---------------------------------------------------------------

	if { $IMCAGui::beamline_name == "imca17bm" } {
		set command "feedback_control bpm_ds &"

		button $button_grid.feedback_button \
			-bg limegreen -activebackground limegreen \
			-text "Feedback Control" \
			-command [ itcl::code eval exec $command ]

		grid $button_grid.feedback_button -row 1 -column $column \
			-padx 5 -pady 5 -sticky nsew
	}

	incr column
}

proc change_edge_energy { record_list entryfield } {
	set edge_desired [ $entryfield get ]

	set record [ $record_list get_record "edge_energy" ]

	$record send_variable $edge_desired
}

proc change_id_ev_enabled { record_list entryfield } {
	set id_ev_enabled_desired [ $entryfield get ]

	set record [ $record_list get_record "id_ev_enabled" ]

	$record send_variable $id_ev_enabled_desired
}

proc change_id_ev_harmonic { record_list entryfield } {
	set id_ev_harmonic_desired [ $entryfield get ]

	set record [ $record_list get_record "id_ev_harmonic" ]

	$record send_variable $id_ev_harmonic_desired
}

proc change_id_ev_params { record_list entryfield } {
	set id_ev_params_desired [ $entryfield get ]

	set record [ $record_list get_record "id_ev_params" ]

	$record send_variable $id_ev_params_desired
}

proc staff_window_callback { } {

	global callback_values

	set window $IMCAGui::staff_window

	set edge_energy $callback_values(edge_energy)

	set edge_energy_value_window $window.grid.edge_value

	$edge_energy_value_window configure -text $edge_energy

	#--------------------------------------------------------------

	if { $IMCAGui::beamline_name == "imca17id" } {
		set id_ev_enabled $callback_values(id_ev_enabled)

		set id_ev_enabled_value_window $window.grid.id_ev_enabled_value

		$id_ev_enabled_value_window configure -text $id_ev_enabled

		#--------------------------------------------------------------

		set id_ev_harmonic $callback_values(id_ev_harmonic)

		set id_ev_harmonic_value_window \
					$window.grid.id_ev_harmonic_value

		$id_ev_harmonic_value_window configure -text $id_ev_harmonic

		#--------------------------------------------------------------

		set id_ev_params $callback_values(id_ev_params)

		set id_ev_params_value_window $window.grid.id_ev_params_value

		$id_ev_params_value_window configure -text $id_ev_params
	}
}

############################################################################

proc create_help_menu { menubar } {
	menu $menubar.help -tearoff 0

	$menubar add cascade -label Help -menu $menubar.help

	$menubar.help add command -label "About IMCA GUI" \
			-command menubar_about_command

	$menubar.help add command -label "Help for IMCA 17-ID" \
			-command [ itcl::code display_help "imca17id.html" ]

	$menubar.help add command -label "Help for IMCA 17-BM" \
			-command [ itcl::code display_help "imca17bm.html" ]

	$menubar.help add command -label "IMCA FAQ" \
			-command [ itcl::code display_help "imca_faq.html" ]

	$menubar.help add command -label "IMCA-CAT FAQ" \
			-command [ itcl::code display_help "imca_cat_faq.html" ]

	proc menubar_about_command {} {
		set t "IMCA GUI was written by William Lavender based on "
		append t "a design by the IMCA-CAT staff."

		tk_dialog .about_dialog "About IMCA GUI" $t info 0 "Done"
	}
}

