#
#  Copyright 1999, 2002, 2004 Illinois Institute of Technology
#
#  See the file "LICENSE" for information on usage and redistribution
#  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#

proc calc_fvalues { toplevel } {

	set element_name_widget $toplevel.main.top_block.element_name
	set edge_selection_widget $toplevel.main.top_block.edge_selection
	set scan_type_widget $toplevel.main.middle_block.omenu

	set element [ $element_name_widget get ]

	if { $element == "" } {
		set t "No element has been selected yet."

		Mx::error_message "No Element Selected" $t "" ""

		return
	}

	set edge [ $edge_selection_widget get ]

	if { $edge == "" } {
		set t "No edge has been selected for the element $element yet."

		Mx::error_message "No Edge Selected" $t error 0 "OK"

		return
	}

	set scan_type [ $scan_type_widget get ]

	set chooch_filename_widget \
		$IMCAGui::experiment_window.bottom.input.chooch_filename

	set filename [ $chooch_filename_widget get ]

	if { $filename == "" } {
		set t "No input scan for Chooch has been selected yet."

		Mx::error_message "No Scan Selected" $t "" ""

		return
	}

	set directory $IMCAGui::output_directory

	if { $directory == "" } {
		set t "No output directory has been set yet."

		Mx::error_message "No Directory Set" $t "" ""

		return
	}

	catch { file delete ${filename}.efs }
	catch { file delete ${filename}.inf }
	catch { file delete ${filename}.ps }

	if { $scan_type == "Absorption" } {
		set script "chooch_xafs.py"
	} else {
		set script "chooch_fluorescence.py"
	}

	set command "xterm -sb -T Chooch -n Chooch -e chooch.wrapper "
	append command "$script $element $edge $filename $directory &"

	set chooch_pid [ eval exec $command ]

	log_puts \
	"Running Chooch for the $element $edge edge.  PID = $chooch_pid"

	after 1000 [ itcl::code is_chooch_running $toplevel $chooch_pid \
			$element $edge $filename $directory ]
}

proc is_chooch_running { toplevel chooch_pid \
				element edge filename directory } {

#	set status [ eval exec "process_state $chooch_pid" ]

	set status [ Mx::process_exists $chooch_pid ]

#	log_puts "Chooch status = $status"

	if { $status == 0 } {
		after 1 \
    [ itcl::code chooch_is_done $toplevel $element $edge $filename $directory ]

		return
	}
	if { $status == 1 } {
		after 1000 \
			[ itcl::code is_chooch_running $toplevel $chooch_pid \
			$element $edge $filename $directory ]

		return
	}

	set t "An unexpected error occurred while checking "
	append t "the state of the Chooch process.  "
	append t "Error text = '$status'"

	Mx::error_message "Error Monitoring Chooch" $t "" ""

	return
}

proc chooch_is_done { toplevel element edge filename directory } {

	log_puts "Chooch is done."

	read_chooch_efs_file $filename.efs

	chooch_update_mad_display $toplevel

	return
}

proc read_chooch_efs_file { efs_filename } {

	set efs [ open $efs_filename r ]

	# Skip the first line.  It is blank.

	set line_number 1

	gets $efs

	# The next line contains the number of measurements in the file.

	incr line_number

	gets $efs line

	if { [ scan $line {%d} num_measurements ] != 1 } {
		return -code error -errorinfo \
		"Could not read number of measurements on line $line_number"
	}

	# Now loop reading the rest of the file.

	set peak_energy -1
	set edge_energy -1

	set peak_value -1000000000000.0
	set edge_value  1000000000000.0

	incr line_number

	gets $efs line

	while { [ eof $efs ] == 0 } {

		if { [ scan $line {%g %g %g} energy fpp fp ] != 3 } {
			return -code error -errorinfo \
		"Error parsing efs file line $line_number.  Line = '$line'"
		}

		set IMCAGui::chooch_fp_array($energy) $fp
		set IMCAGui::chooch_fpp_array($energy) $fpp

		if { $fpp > $peak_value } {

			set peak_value $fpp

			set peak_energy $energy
		}
		if { $fp < $edge_value } {

			set edge_value $fp

			set edge_energy $energy
		}

		incr line_number

		gets $efs line
	}

	close $efs

	set IMCAGui::chooch_peak $peak_energy
	set IMCAGui::chooch_edge $edge_energy

	log_puts "Edge at $edge_energy, fp = $edge_value"
	log_puts "Peak at $peak_energy, fpp = $peak_value"

	return
}

proc chooch_update_mad_display { toplevel } {

	set grid $toplevel.bottom.grid

	# Update peak values

	set energy $IMCAGui::chooch_peak

	$grid.peak_energy configure -text $energy

	set wavelength [ lambda_from_energy $energy ]

	set wavelength [ format {%.7g} $wavelength ]

	$grid.peak_wavelength configure -text $wavelength

	$grid.peak_fprime configure -text $IMCAGui::chooch_fp_array($energy)

	$grid.peak_fdblprime configure -text $IMCAGui::chooch_fpp_array($energy)

	# Update edge values

	set energy $IMCAGui::chooch_edge

	$grid.edge_energy configure -text $energy

	set wavelength [ lambda_from_energy $energy ]

	set wavelength [ format {%.7g} $wavelength ]

	$grid.edge_wavelength configure -text $wavelength

	$grid.edge_fprime configure -text $IMCAGui::chooch_fp_array($energy)

	$grid.edge_fdblprime configure -text $IMCAGui::chooch_fpp_array($energy)

	return
}

proc chooch_display_far_params { toplevel location type } {

	if { $IMCAGui::chooch_peak == "" } {
		set t \
"Can't compute parameters for this position since Chooch has not been run yet."

		Mx::error_message "Chooch has not been run yet" $t "" ""

		return
	}

	set grid $toplevel.bottom.grid

	if { $type == "wavelength" } {
		set wavelength [ $grid.${location}_wavelength get ]

		set energy [ energy_from_lambda $wavelength ]

		$grid.${location}_energy clear
		$grid.${location}_energy insert 0 $energy
	} else {
		set energy [ $grid.${location}_energy get ]

		set wavelength [ lambda_from_energy $energy ]

		$grid.${location}_wavelength clear
		$grid.${location}_wavelength insert 0 $wavelength
	}

	set energy_interval [ find_chooch_energy_interval $energy ]

	if { [ llength $energy_interval ] == 1 } {

		set selected_energy [ lindex $energy_interval 0 ]

		if { $selected_energy == -1 } {

			set fp_value "Out of range"
			set fpp_value "Out of range"
		} else {
			set fp_value \
				$IMCAGui::chooch_fp_array($selected_energy)
			set fpp_value \
				$IMCAGui::chooch_fpp_array($selected_energy)
		}
	} else {
		set fp_value \
			[ interpolate_fvalue $energy $energy_interval "fp" ]

		set fpp_value \
			[ interpolate_fvalue $energy $energy_interval "fpp" ]
	}

	$grid.${location}_fprime configure -text $fp_value
	$grid.${location}_fdblprime configure -text $fpp_value

	return
}

proc find_chooch_energy_interval { energy } {

	set threshold 0.001

	set energy_list [ array names IMCAGui::chooch_fp_array ]

	set energy_list [ lsort -real $energy_list ]

	set first_energy [ lindex $energy_list 0 ]

	if { $energy < ( $first_energy - $threshold ) } {

		return [ list -1 ]
	}

	set list_length [ llength $energy_list ]

	set end_marker [ expr ( $list_length - 1 ) ]

	for { set i 0 } { $i < $end_marker } { incr i } {

		set current_energy [ lindex $energy_list $i ]

		if { abs( $energy - $current_energy ) < $threshold } {

			return [ list $current_energy ]
		}

		set j [ expr ( $i + 1 ) ]

		set next_energy [ lindex $energy_list $j ]

		if { $energy < $next_energy } {

			return [ list $current_energy $next_energy ]
		}
	}
	return [ list -1 ]
}

proc interpolate_fvalue { energy energy_interval type } {

	set low_energy [ lindex $energy_interval 0 ]
	set high_energy [ lindex $energy_interval 1 ]

	if { $type == "fp" } {
		set low_fvalue $IMCAGui::chooch_fp_array($low_energy)
		set high_fvalue $IMCAGui::chooch_fp_array($high_energy)
	} else {
		set low_fvalue $IMCAGui::chooch_fpp_array($low_energy)
		set high_fvalue $IMCAGui::chooch_fpp_array($high_energy)
	}

	set denominator [ expr ( $high_energy - $low_energy ) ]

	if { abs($denominator) < 1.0e-30 } {
		return "Energy scan steps too small"
	}

	set numerator [ expr ( $high_fvalue - $low_fvalue ) ]

	set slope [ expr ( $numerator / $denominator ) ]

	set intercept [ expr ( $low_fvalue - $slope * $low_energy ) ]

	set fvalue [ expr ( $slope * $energy + $intercept ) ]

	set fvalue [ format {%.4g} $fvalue ]

	return $fvalue
}

proc print_chooch_plot { } {

	global tcl_platform

	set chooch_filename_widget \
		$IMCAGui::experiment_window.bottom.input.chooch_filename

	set filename [ $chooch_filename_widget get ]

	if { $filename == "" } {
		set t "No input scan for Chooch has been selected yet."

		Mx::error_message "No Scan Selected" $t "" ""

		return
	}

	set print_filename "$filename.ps"

	log_puts "Printing $print_filename"

	set platform "$tcl_platform(os)"

	if { $platform == "IRIX" } {
		set command "lp $print_filename &"
	} else {
		set command "lpr $print_filename &"
	}

	eval exec $command
}

