#
#  cryst_func.tcl -- Generic crystallography program support functions.
#
#  Written by William Lavender.
#
#------------------------------------------------------------------------
#
#  Copyright 1999, 2002, 2004-2005 Illinois Institute of Technology
#
#  See the file "LICENSE" for information on usage and redistribution
#  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#

proc create_message_log_window { label } {

	set window .message_log

	frame $window
	pack $window -expand yes -fill both

	frame $window.top
	pack $window.top -side top -expand yes -fill x

	if { $IMCAGui::use_tabs == 0 } {
		label $window.top.label -text "Message Log"
		pack $window.top.label -side top
	} else {
		label $window.top.left_filler -text "           "
		label $window.top.label -text "Message Log"

		button $window.top.exit_button -text "Exit Imcagui" \
			-bg lightblue -activebackground lightblue \
			-command exit

		grid $window.top.left_filler -row 0 -column 0 -padx 20 -sticky w
		grid $window.top.label       -row 0 -column 1 -padx 10
		grid $window.top.exit_button -row 0 -column 2 -padx 10 -sticky e

		grid columnconfigure $window.top 0 -weight 1
		grid columnconfigure $window.top 1 -weight 1
		grid columnconfigure $window.top 2 -weight 1
	}

        #
        #  Create the log window.
        #

        iwidgets::scrolledtext $window.log

        $window.log configure -visibleitems 56x10
        $window.log configure -hscrollmode dynamic
        $window.log configure -vscrollmode dynamic

        pack $window.log -expand yes -fill both -side top

        #
        #  Install handlers to redirect the output of mx_info(), mx_error()
        #  and MX_DEBUG() to this window.
        #

	Mx::set_all_output_functions imcagui_output_function

	#
	#  Install handler for mx_info_dialog().
	#

	Mx::set_info_dialog_function imcagui_info_dialog_function

	log_puts "*** $label Initializing ***"
}

proc imcagui_output_function { message } {
	log_puts $message
	update
}

proc imcagui_info_dialog_function { text_prompt gui_prompt button_label } {
	bell
	tk_dialog .imcagui_info_dialog "MX Info" \
		$gui_prompt info 0 $button_label
}

proc imcagui_user_interrupt_function {} {

	update

	set interrupt_occurred $IMCAGui::user_interrupt

	set IMCAGui::user_interrupt 0

	return $interrupt_occurred
}

proc log_puts { message } {
	set textwindow [ .message_log.log childsite ]

	$textwindow configure -state normal

	.message_log.log insert end "$message\n"

	$textwindow configure -state disabled

	.message_log.log yview moveto 1.0
}

proc energy_from_lambda { lambda } {

	set status [ catch { expr 12398.4186 / $lambda } energy ]

	if { $status != 0 } {
		set energy [ expr 1.0e308 ]
	}
	return $energy
}

proc lambda_from_energy { energy } {

	set status [ catch { expr 12398.4186 / $energy } lambda ]

	if { $status != 0 } {
		set lambda [ expr 1.0e308 ]
	}
	return $lambda
}

proc print_window { window } {

	set id [ winfo id $window ]

	set command "printwindow $id &"

	eval exec $command
}

proc set_default_printer {} {

	global env

	if [ catch { set default_printer $env(PRINTER) } ] {
		set default_printer ""
	}

	iwidgets::promptdialog .set_default_printer_dialog

	.set_default_printer_dialog configure -title "Set Default Printer"
	.set_default_printer_dialog configure -labeltext "Default Printer"
	.set_default_printer_dialog hide Apply

	.set_default_printer_dialog buttonconfigure OK -command \
					set_default_printer_ok
	.set_default_printer_dialog buttonconfigure Cancel -command \
					set_default_printer_cancel

	.set_default_printer_dialog insert 0 $default_printer

	.set_default_printer_dialog activate
}

proc set_default_printer_ok {} {

	global env

	set default_printer [ .set_default_printer_dialog get ]

	set default_printer [ string trim $default_printer ]

	if { $default_printer == "" } {
		catch { unset env(PRINTER) }
	} else {
		set env(PRINTER) $default_printer
	}

	.set_default_printer_dialog deactivate
	destroy .set_default_printer_dialog

	log_puts "Setting default printer to '$default_printer'."
}

proc set_default_printer_cancel {} {
	.set_default_printer_dialog deactivate
	destroy .set_default_printer_dialog
}

set log_puts_exists 1

