#
#  Copyright 1999, 2002-2004 Illinois Institute of Technology
#
#  See the file "LICENSE" for information on usage and redistribution
#  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#

proc create_experiment_window { type record_list toplevel_window } {

	if { $toplevel_window == "" } {

		set toplevel .experiment_window

		set status [ catch { toplevel $toplevel } result ]

		# If the window already exists, bring it to the top.

		if { $status != 0 } {
			raise $toplevel
			focus $toplevel

			select_experiment_type $toplevel $type
			return
		}
		focus $toplevel

		wm title $toplevel "Spectroscopy & MAD Data Collection Setup"

		create_experiment_menubar $toplevel $record_list

		set window $toplevel
	} else {
		set window $toplevel_window
	}

	set IMCAGui::experiment_window $window

	create_experiment_selector $window

	if { $IMCAGui::use_tabs == 0 } {
		frame $window.sep0 -height 2 -borderwidth 1 -relief sunken
		pack $window.sep0 -side top -expand yes -fill x -pady 5
	}

	create_main_spectroscopy_window $window $record_list

	frame $window.sep1 -height 2 -borderwidth 1 -relief sunken
	pack $window.sep1 -side top -expand yes -fill x -pady 5

	create_mad_bottom_window $window $record_list

	select_experiment_type $window $type
}

proc create_experiment_menubar { toplevel record_list } {

	set menubar $toplevel.menubar

	menu $menubar
	$toplevel configure -menu $menubar

	menu $menubar.help -tearoff 0
	$menubar add cascade -label Help -menu $menubar.help

	$menubar.help add command -label "Help for Spectroscopy" \
			-command [ itcl::code display_help "spectroscopy.html" ]
	$menubar.help add command -label "Help for MAD" \
			-command [ itcl::code display_help "mad.html" ]
	$menubar.help add command -label "Help for Chooch" \
			-command [ itcl::code display_help "chooch.html" ]
}

proc create_experiment_selector { toplevel } {

	iwidgets::optionmenu $toplevel.type -labeltext "Experiment Type" \
		-command [ itcl::code select_experiment_command $toplevel ]

	if { $IMCAGui::use_tabs == 0 } {
		pack $toplevel.type
	}

	$toplevel.type insert end Spectroscopy MAD
	$toplevel.type select MAD
}

proc select_experiment_command { toplevel } {

	set type [ $toplevel.type get ]

	select_experiment_type $toplevel $type
}

proc select_experiment_type { toplevel type } {

	switch -exact -- $type {
		Spectroscopy { pack forget $toplevel.bottom; \
			pack forget $toplevel.sep2 }

		MAD { pack $toplevel.bottom -side top -after $toplevel.sep1; \
			pack $toplevel.sep2 -side top -expand yes \
				-fill x -pady 5 -after $toplevel.bottom }

		default { Mx::error_message "Unexpected Experiment Type" \
	"Experiment type should be 'Spectroscopy' or 'MAD', not '$type'" \
			"" "" }
	}

	$toplevel.type select $type
}

