#
#  Copyright 1999, 2002-2004 Illinois Institute of Technology
#
#  See the file "LICENSE" for information on usage and redistribution
#  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#

proc create_mad_bottom_window { toplevel record_list } {

	create_mad_grid $toplevel $record_list

	frame $toplevel.sep2 -height 2 -borderwidth 1 -relief sunken
	pack $toplevel.sep2 -side top -expand yes -fill x -pady 5

	create_mad_buttonrow $toplevel $record_list
}

proc create_mad_grid { toplevel record_list } {

	set bottom $toplevel.bottom

	frame $bottom
	pack $bottom -side top

	set input $bottom.input

	frame $input
	pack $input -side top

	iwidgets::entryfield $input.chooch_filename \
				-labeltext "Input Scan File"
	pack $input.chooch_filename -side left

	button $input.browse -text Browse -command \
				[ itcl::code select_file $input ]
	pack $input.browse -side left

	set button1_frame $bottom.button1_frame

	frame $button1_frame
	pack $button1_frame

	button $button1_frame.calc_fvalues \
		-text "Calculate f', f''" \
		-command [ itcl::code calc_fvalues $toplevel ]

	pack $button1_frame.calc_fvalues -side left

	set grid $bottom.grid

	frame $grid
	pack $grid -side top

	#
	# Please note in the following code that it is intentional
	# that the items in the "High" columns ( columns 8 and 9 ) are not
	# mapped onto the screen.  The value of $grid.num_wavelengths
	# is what determines whether or not the "High" row is displayed.
	#

	set entrywidth 10

	set row 0

	iwidgets::optionmenu $grid.num_wavelengths \
		-labelpos e -labeltext "Wavelengths:" \
		-command [ itcl::code mad_wavelength_select $grid ]

	$grid.num_wavelengths insert end Three Four
	$grid.num_wavelengths select Three

	label $grid.peak_label -text Peak
	label $grid.edge_label -text Edge
	label $grid.low_label  -text Low
	label $grid.high_label -text High

	grid $grid.num_wavelengths -row $row -column 1
	grid $grid.peak_label      -row $row -column 2
	grid $grid.edge_label      -row $row -column 4
	grid $grid.low_label       -row $row -column 6

	#-----

	incr row

	label $grid.wavelength_label      -text "Wavelength:"
	label $grid.peak_wavelength       -text "---" -fg blue
	label $grid.peak_wavelength_units -text "A   "
	label $grid.edge_wavelength       -text "---" -fg blue
	label $grid.edge_wavelength_units -text "A   "

	iwidgets::entryfield $grid.low_wavelength -width $entrywidth
	label $grid.low_wavelength_units  -text "A   "

	$grid.low_wavelength configure -command \
	    [ itcl::code chooch_display_far_params $toplevel low wavelength ]

	iwidgets::entryfield $grid.high_wavelength -width $entrywidth
	label $grid.high_wavelength_units -text "A   "

	$grid.high_wavelength configure -command \
	    [ itcl::code chooch_display_far_params $toplevel high wavelength ]

	grid $grid.wavelength_label       -row $row -column 1 -sticky e
	grid $grid.peak_wavelength        -row $row -column 2
	grid $grid.peak_wavelength_units  -row $row -column 3
	grid $grid.edge_wavelength        -row $row -column 4
	grid $grid.edge_wavelength_units  -row $row -column 5
	grid $grid.low_wavelength         -row $row -column 6
	grid $grid.low_wavelength_units   -row $row -column 7

	#-----

	incr row

	label $grid.energy_label      -text "Energy:"
	label $grid.peak_energy       -text "---" -fg blue
	label $grid.peak_energy_units -text "eV  "
	label $grid.edge_energy       -text "---" -fg blue
	label $grid.edge_energy_units -text "eV  "

	iwidgets::entryfield $grid.low_energy -width $entrywidth
	label $grid.low_energy_units  -text "eV  "

	$grid.low_energy configure -command \
		[ itcl::code chooch_display_far_params $toplevel low energy ]

	iwidgets::entryfield $grid.high_energy -width $entrywidth
	label $grid.high_energy_units -text "eV  "

	$grid.high_energy configure -command \
		[ itcl::code chooch_display_far_params $toplevel high energy ]

	grid $grid.energy_label       -row $row -column 1 -sticky e
	grid $grid.peak_energy        -row $row -column 2
	grid $grid.peak_energy_units  -row $row -column 3
	grid $grid.edge_energy        -row $row -column 4
	grid $grid.edge_energy_units  -row $row -column 5
	grid $grid.low_energy         -row $row -column 6
	grid $grid.low_energy_units   -row $row -column 7

	#-----

	incr row

	label $grid.fprime_label      -text "f':"
	label $grid.peak_fprime       -text "---" -fg blue
	label $grid.edge_fprime       -text "---" -fg blue
	label $grid.low_fprime        -text "---" -fg blue
	label $grid.high_fprime       -text "---" -fg blue

	grid $grid.fprime_label       -row $row -column 1 -sticky e
	grid $grid.peak_fprime        -row $row -column 2
	grid $grid.edge_fprime        -row $row -column 4
	grid $grid.low_fprime         -row $row -column 6

	#-----

	incr row

	label $grid.fdblprime_label      -text "f'':"
	label $grid.peak_fdblprime       -text "---" -fg blue
	label $grid.edge_fdblprime       -text "---" -fg blue
	label $grid.low_fdblprime        -text "---" -fg blue
	label $grid.high_fdblprime       -text "---" -fg blue

	grid $grid.fdblprime_label       -row $row -column 1 -sticky e
	grid $grid.peak_fdblprime        -row $row -column 2
	grid $grid.edge_fdblprime        -row $row -column 4
	grid $grid.low_fdblprime         -row $row -column 6

	#-----

	frame $bottom.button_sep -height 2 -borderwidth 1
	pack $bottom.button_sep -side top -expand yes -fill x -pady 1

	set button2_frame $bottom.button2_frame

	frame $button2_frame
	pack $button2_frame

	button $button2_frame.print_chooch_plot \
		-text "Print f', f'' Spectrum" \
		-command print_chooch_plot

	pack $button2_frame.print_chooch_plot
}

proc mad_wavelength_select { grid } {

	set selection [ $grid.num_wavelengths get ]

	switch -exact -- $selection {
		Three { mad_wavelength_hide_high_column $grid }

		Four { mad_wavelength_show_high_column $grid }

		default { Mx::error_message "Unexpected number of wavelengths" \
		"Should have 'Three' or 'Four' wavelengths, not '$selection'" \
			"" "" }
	}
}

proc mad_wavelength_hide_high_column { grid } {

	grid forget $grid.high_label
	grid forget $grid.high_wavelength
	grid forget $grid.high_wavelength_units
	grid forget $grid.high_energy
	grid forget $grid.high_energy_units
	grid forget $grid.high_fprime
	grid forget $grid.high_fdblprime

}

proc mad_wavelength_show_high_column { grid } {

	grid $grid.high_label            -row 0 -column 8
	grid $grid.high_wavelength       -row 1 -column 8
	grid $grid.high_wavelength_units -row 1 -column 9
	grid $grid.high_energy           -row 2 -column 8
	grid $grid.high_energy_units     -row 2 -column 9
	grid $grid.high_fprime           -row 3 -column 8
	grid $grid.high_fdblprime        -row 4 -column 8

}

proc create_mad_buttonrow { toplevel record_list } {

	set buttonrow $toplevel.buttonrow

	frame $buttonrow
	pack $buttonrow -side top

	button $buttonrow.print -text "Print\nWindow" \
				-command [ itcl::code print_window $toplevel ]

	pack $buttonrow.print -side left

	if { $IMCAGui::use_tabs == 0 } {
		frame $buttonrow.sep1 -width 1 -borderwidth 1
		pack $buttonrow.sep1 -side left -expand yes -fill y -padx 20

		button $buttonrow.close -text Close \
				-command [ itcl::code destroy $toplevel ]

		pack $buttonrow.close -side left
	}
}

proc select_file { frame } {

	set dialog .scan_file_dialog

	set selectionbox $dialog.selectionbox
	set selection $selectionbox.selection
	set chooch_filename_widget \
		$IMCAGui::experiment_window.bottom.input.chooch_filename

	set old_filename [ $chooch_filename_widget get ]

	set directory [ file dirname $old_filename ]

	if [ winfo exists $dialog ] {
		$selectionbox configure -directory $directory

		$selection clear
		$selection insert 0 $old_filename

		wm deiconify $dialog
		raise $dialog
		focus $dialog
		return
	}

	toplevel $dialog

	wm title $dialog "Select Input Scan"

	iwidgets::fileselectionbox $selectionbox
	pack $selectionbox

	$selectionbox configure -directory $directory

	$selection clear
	$selection insert 0 $old_filename

	set buttonbox $dialog.buttonbox

	iwidgets::buttonbox $buttonbox
	pack $buttonbox

	$buttonbox add OK -text OK -command \
		[ itcl::code select_file_ok $dialog $chooch_filename_widget ]

	$buttonbox add Cancel -text Cancel -command \
		[ itcl::code select_file_cancel $dialog ]

	$buttonbox default OK

	wm deiconify $dialog
}

proc select_file_ok { dialog chooch_filename_widget } {

	set filename [ $dialog.selectionbox get ]

	$chooch_filename_widget clear
	$chooch_filename_widget insert end $filename

	wm withdraw $dialog
}

proc select_file_cancel { dialog } {

	wm withdraw $dialog
}

